#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# simple RVE with 2 materials for testing periodic BCs

import heat as h
import tbox
import tbox.gmsh as gmsh

def main():

    msh = gmsh.MeshLoader("rve.geo", __file__).execute()

    pbl = h.Problem(msh)

    solver = h.Solver(pbl)
    solver.nthreads = 4

    h.Medium(pbl, "Matrix",    tbox.Fct2C(10.,1.,0), 1.0)
    h.Medium(pbl, "Inclusion", tbox.Fct2C(100.,100.,0), 1.0)

    TM   = 55.0
    dTxM = 30.0
    dTyM = 1.0
    h.Periodic(pbl, TM, dTxM, dTyM)

    solver.start()

    from heat.viewer import GUI
    GUI().open(pbl.msh.name)
    
if __name__ == "__main__":
    main()
