# -*- coding: utf-8 -*-
# test encoding: à-é-è-ô-ï-€


# Copyright 2022 University of Liège
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import sys
import waves as w
import tbox as tb

# TODO: creer un objet dataset qui gere l'ugrid (comme dans Metafor)

# Qt
try:
    from PyQt5.QtCore import *
    from PyQt5.QtGui import *
    from PyQt5.QtWidgets import *
    print("PyQt5 (Qt %s) loaded!" % QT_VERSION_STR)
except :
    try:
        import  PySide2
        from PySide2.QtCore import *
        from PySide2.QtGui import *
        from PySide2.QtWidgets import *
        #print("PySide2  loaded!")
        print("PySide2 (Qt %s) loaded!" % PySide2.__version__)
    except :
        print("Unable to load PyQt nor PySide2")
        sys.exit()
        

# vtk
import vtk
try:
    from vtk.qt.QVTKRenderWindowInteractor import QVTKRenderWindowInteractor
    print("using vtk.qt.QVTKRenderWindowInteractor")
except:
    from tboxVtk.qt5vtk6.QVTKRenderWindowInteractor import QVTKRenderWindowInteractor
    print("using pki.qt5vtk6.QVTKRenderWindowInteractor")

# -----------------------------------------------------------------------------------------


class MeshViewer(QWidget, w.DisplayHook):

    app = QApplication(sys.argv)
    """
    Qt GUI for visu. the output
    """

    def __init__(self, scheme):
        #super(MeshViewer, self).__init__() # a eclaircir... marche pô
        w.DisplayHook.__init__(self)
        QWidget.__init__(self)

        scheme.setGUI(self)

        self.msh = scheme.pbl.msh
        self.scheme = scheme
        self.writer = tb.GmshExport(self.msh)

        self.running = 'init'
        self.vmin = -0.1
        self.vmax = 0.1

        print("starting MeshViewer init...")

        self.__setupGUI()
        self.__setupVTK()

        self.app.lastWindowClosed.connect(self.app.quit)
        self.show()
        self.vtkwidget.Initialize(
        )  # [Linux] faire ca le plus tard possible (apres self.show!!)
        print("MeshViewer ready.")

    def closeEvent(self, event):
        self.scheme.stopit = True
        self.running = 'running'  # sort de "while self.running=='pause'"
        print("GUI killed!")
        QWidget.closeEvent(self, event)

    def start(self):
        self.app.exec_()

    def __setupGUI(self):

        self.setWindowTitle("MeshViewer")
        self.resize(800, 600)

        # vtk window

        self.vtkwidget = QVTKRenderWindowInteractor(
            self)  # "self" sinon, rien ne s'affiche

        self.vtkwidget.setSizePolicy(
            QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding))
        self.vtkwidget.setMinimumSize(QSize(300, 300))
        self.vtkwidget.setSizePolicy(QSizePolicy.Expanding,
                                     QSizePolicy.Expanding)

        #self.vtkwidget.Initialize() # plante sous Linux!!

        self.renderer = vtk.vtkRenderer()
        self.renderer.SetBackground(1.0, 1.0, 1.0)
        self.vtkwidget.GetRenderWindow().AddRenderer(self.renderer)

        style = vtk.vtkInteractorStyleTrackballCamera()
        self.vtkwidget.SetInteractorStyle(style)

        hbox = QHBoxLayout()
        self.setLayout(hbox)
        hbox.addWidget(self.vtkwidget)

        pan = QFrame()
        pan.setMaximumSize(QSize(200, 999999))
        hbox.addWidget(pan)
        #hbox.setMargin(0)    # sinon vtkwidget degouline en bas
        #hbox.setSpacing(0)   # sinon vtkwidget clignote sur le côté

        vbox = QVBoxLayout()
        pan.setLayout(vbox)

        self.startBut = QPushButton(self.tr("start!"))
        self.startBut.clicked.connect(self.startSlot)
        vbox.addWidget(self.startBut)

        groupBox = QGroupBox("Scale")
        gbox = QGridLayout()
        label = QLabel("vmin")
        gbox.addWidget(label, 0, 0)
        self.vminlinedt = QLineEdit('%f' % self.vmin)
        self.vminlinedt.editingFinished.connect(self.updateTCoords)
        gbox.addWidget(self.vminlinedt, 0, 1)
        label = QLabel("vmax")
        gbox.addWidget(label, 1, 0)
        self.vmaxlinedt = QLineEdit('%f' % self.vmax)
        self.vmaxlinedt.editingFinished.connect(self.updateTCoords)
        gbox.addWidget(self.vmaxlinedt, 1, 1)
        groupBox.setLayout(gbox)
        vbox.addWidget(groupBox)

        verbchk = QCheckBox("verbose")
        verbchk.setChecked(self.scheme.verbose)
        verbchk.stateChanged.connect(self.verboseSlot)
        vbox.addWidget(verbchk)

        groupBox = QGroupBox("Infos")
        self.steplabel = QLabel("step # 0")
        self.timelabel = QLabel("time = 0")
        gbox = QVBoxLayout()
        groupBox.setLayout(gbox)
        gbox.addWidget(self.steplabel)
        gbox.addWidget(self.timelabel)
        vbox.addWidget(groupBox)

        vbox.addStretch(1)

    def verboseSlot(self, state):
        self.scheme.verbose = (state != 0)
        if not state:
            print("<verbose=OFF>")

    def startSlot(self):
        if self.running == 'init':
            self.startBut.setText("Pause")  # on demarre et on affiche "pause"
            self.running = 'running'
            self.scheme.start(self.writer)
            self.startBut.setText("Quit")
            self.running = 'quit'
        elif self.running == 'running':  # on stoppe et on affiche 'continue"
            self.running = 'pause'
            self.startBut.setText("Continue")
            while self.running == 'pause':
                self.app.processEvents(QEventLoop.WaitForMoreEvents)
        elif self.running == 'pause':
            self.running = 'running'
            self.startBut.setText("Pause")
        elif self.running == 'quit':
            self.app.quit()

    def disableStart(self):
        self.startBut.setDisabled(True)

    def __setupVTK(self):

        self.__createUgrid()

        # mesh
        self.mesh = Mesh(self.ugrid)
        self.renderer.AddActor(self.mesh.actor)

        # grid
        self.grid = Grid(self.ugrid)
        self.renderer.AddViewProp(self.grid.actor)

        #scalarbar
        self.scalarbar = ScalarBar(self.mesh.lut)
        self.renderer.AddViewProp(self.scalarbar.actor)

        self.resetCamera()

    def resetCamera(self):
        self.renderer.ResetCamera()
        cam1 = self.renderer.GetActiveCamera()
        # 3D
        if 0:
            cam1.Elevation(35)
            cam1.SetViewUp(0, 1, 0)
            cam1.Azimuth(30)
        #2D
        else:
            cam1.Elevation(0)
            cam1.SetViewUp(0, 1, 0)
            cam1.Azimuth(0)
        self.renderer.ResetCameraClippingRange()

    def display(self, nt, t, u):
        """ C++ => python callback
        """
        #print "[GUI] nt=", nt, "t=", t

        self.steplabel.setText("step # %d (%d%%)" %
                               (nt, t / self.scheme.ttot * 100))
        self.timelabel.setText("time = %f" % t)

        # update data at points
        for i in range(self.msh.nodes.size()):
            self.scalars.SetValue(i, u[i])
        self.scalars.Modified()

        # update texture coords
        #vmin, vmax = self.scalars.GetValueRange()
        #vmin=-0.01
        #vmax=0.01

        self.updateTCoords()

    def updateTCoords(self):
        try:
            self.vmin = float(self.vminlinedt.text())
            self.vmax = float(self.vmaxlinedt.text())
        except:
            pass

        self.mesh.updateTcoords(self.vmin, self.vmax)
        self.render()

    def ragequit(self):
        print("rage quit!")
        self.scheme.stopit = True
        self.app.quit()

        #raise Exception('GUI killed!')

    def render(self):
        # draw the scene
        self.vtkwidget.Render()
        self.app.processEvents()

    def refresh(self):
        #print "[GUI] refresh"
        self.app.processEvents()

    def __createUgrid(self):
        print('creating vtkUnstructuredGrid...')
        self.ugrid = vtk.vtkUnstructuredGrid()
        self.points = vtk.vtkPoints()
        self.ugrid.SetPoints(self.points)

        self.scalars = vtk.vtkFloatArray()
        self.scalars.SetNumberOfComponents(1)
        self.ugrid.GetPointData().SetScalars(self.scalars)

        # points / data
        for i in range(self.msh.nodes.size()):
            p = self.msh.nodes[i].pos
            self.points.InsertPoint(i, p[0], p[1], p[2])
            #val = p.x()*p.x()+p.y()*p.y()+p.z()*p.z()
            val = 0.0
            self.scalars.InsertNextValue(val)

        # hexa
        for i in range(self.msh.elems.size()):
            e = self.msh.elems[i]
            n = e.nodes
            if e.type() == tb.ELTYPE_HEX8:
                hexa = vtk.vtkHexahedron()
                ids = hexa.GetPointIds()
                for j in range(8):
                    ids.SetId(j, n[j].no - 1)
                self.ugrid.InsertNextCell(hexa.GetCellType(), ids)
        print(self.ugrid.GetNumberOfPoints(
        ), 'points and', self.ugrid.GetNumberOfCells(), 'cells converted')

    def writeXML(self, name='mesh.vtu'):
        writer = vtk.vtkXMLUnstructuredGridWriter()
        compressor = vtk.vtkZLibDataCompressor()
        writer.SetCompressor(compressor)
        writer.SetDataModeToAscii()
        writer.SetInputConnection(self.producer.GetOutputPort())
        writer.SetFileName(name)
        writer.Write()

    def write(self, name='mesh.vtk'):
        writer = vtk.vtkUnstructuredGridWriter()
        writer.SetFileName(name)
        writer.SetFileTypeToASCII()
        writer.SetInputConnection(self.producer.GetOutputPort())
        writer.Write()


class Mesh:
    def __init__(self, ugrid):

        self.ugrid = ugrid

        # compatibility vtk 5/6 => use a trivial producer
        #self.producer = vtk.vtkTrivialProducer()
        #self.producer.SetOutput(self.ugrid) # le double producer (dans Mesh/Grid) ne marche pas sous Qt 4 => faire un objet dataset

        self.__createTexture()
        self.__createTcoords()
        vmin, vmax = self.ugrid.GetPointData().GetScalars().GetValueRange()
        self.updateTcoords(vmin, vmax)

        self.mapper = vtk.vtkDataSetMapper()
        #self.mapper.SetInputConnection(self.producer.GetOutputPort())
        self.mapper.SetInputData(ugrid)
        if vtk.VTK_MAJOR_VERSION <= 7:
            self.mapper.ImmediateModeRenderingOff()
        self.mapper.SetResolveCoincidentTopologyToPolygonOffset()
        self.mapper.ScalarVisibilityOff()

        self.actor = vtk.vtkActor()
        self.actor.SetMapper(self.mapper)
        self.actor.SetTexture(self.texture)

    def __createTcoords(self):
        self.tcoords = vtk.vtkFloatArray()
        self.tcoords.SetNumberOfComponents(2)
        self.tcoords.Resize(
            self.ugrid.GetPointData().GetScalars().GetNumberOfTuples())
        self.ugrid.GetPointData().SetTCoords(self.tcoords)

    def updateTcoords(self, vmin, vmax):
        #print 'vmin=',vmin,' vmax=',vmax
        if vmin == vmax:
            dist = 1.0
        else:
            dist = vmax - vmin
        thr = 1. / 256.
        scalars = self.ugrid.GetPointData().GetScalars()
        for i in range(scalars.GetNumberOfTuples()):
            val = scalars.GetTuple1(i)
            tc = (val - vmin) / dist
            if (tc < thr): tc = thr
            if (tc > 1.0 - thr): tc = 1.0 - thr
            self.tcoords.InsertTuple2(i, tc, tc)

        self.lut.SetTableRange(vmin, vmax)  # utile pour la scalarbar seulement
        self.lut.Build()

    def __createTexture(self):
        # lookup table
        #(utilisée 1x au demarrage pour la texture)
        #(puis pour la scalarbar)
        self.lut = vtk.vtkLookupTable()
        self.lut.SetNumberOfColors(16)
        self.lut.SetHueRange(0., 0.667)
        self.lut.SetTableRange(
            0.0, 1.0)  # valeurs bidons -> pas d'infl sur la texture
        self.lut.Build()

        self.texture = vtk.vtkTexture()
        dim = self.lut.GetNumberOfTableValues()
        img = vtk.vtkImageData()
        img.SetDimensions(dim, dim, 1)

        if vtk.VTK_MAJOR_VERSION <= 5:
            img.SetScalarTypeToUnsignedChar()
            img.SetNumberOfScalarComponents(3)
            img.AllocateScalars()
        else:
            img.AllocateScalars(vtk.VTK_UNSIGNED_CHAR, 3)

        for i in range(dim):
            ci = self.lut.GetTableValue(i)
            for k in range(3):
                img.SetScalarComponentFromDouble(i, i, 0, k, ci[k] * 255)
        if vtk.VTK_MAJOR_VERSION <= 5:
            self.texture.SetInput(img)
        else:
            self.texture.SetInputData(img)

        if 0:  # --display texture using vtkImageViewer2 (DEBUG)
            imageViewer = vtk.vtkImageViewer2()
            if vtk.VTK_MAJOR_VERSION <= 5:
                imageViewer.SetInput(img)
            else:
                imageViewer.SetInputData(img)
            renderWindowInteractor = vtk.vtkRenderWindowInteractor()
            imageViewer.SetupInteractor(renderWindowInteractor)
            imageViewer.Render()
            imageViewer.GetRenderer().ResetCamera()
            imageViewer.Render()
            renderWindowInteractor.Start()


class ScalarBar:
    def __init__(self, lut):
        self.actor = vtk.vtkScalarBarActor()
        self.actor.SetLookupTable(lut)
        self.actor.GetPositionCoordinate(
        ).SetCoordinateSystemToNormalizedViewport()
        self.actor.GetPositionCoordinate().SetValue(0.1, 0.05)
        self.actor.SetOrientationToHorizontal()
        self.actor.SetWidth(0.8)
        self.actor.SetHeight(0.1)
        self.actor.GetTitleTextProperty().SetFontFamilyToArial()
        self.actor.GetLabelTextProperty().SetFontFamilyToArial()
        self.actor.SetLabelFormat("%g")
        self.actor.GetTitleTextProperty().SetColor(0., 0., 0.)
        self.actor.GetLabelTextProperty().SetColor(0., 0., 0.)
        self.actor.SetTitle("pressure")


class Grid:
    def __init__(self, ugrid):
        self.mapper = vtk.vtkDataSetMapper()
        if vtk.VTK_MAJOR_VERSION <= 7:
            self.mapper.ImmediateModeRenderingOff()
        self.mapper.SetResolveCoincidentTopologyToPolygonOffset()
        self.mapper.ScalarVisibilityOff()
        self.mapper.SetInputData(ugrid)
        self.actor = vtk.vtkActor()

        self.actor.GetProperty().SetRepresentationToWireframe()
        self.actor.GetProperty().SetColor(0., 0., 0.)
        self.actor.GetProperty().SetAmbient(1.0)
        self.actor.GetProperty().SetDiffuse(0.0)
        self.actor.GetProperty().SetSpecular(0.0)
        self.actor.SetMapper(self.mapper)
